#!/usr/bin/env python3

__author__ = 'Martin'

import tree_utils

# from Wikipedia pseudocode

###############################################################################################################
# Repair the heap whose root element is at index 'start',
# assuming the heaps rooted at its children are valid
#
def siftDown(a, start, end):
    print('\nsiftDown in', a)
    root = start

    while root * 2 + 1 <= end:     # While the root has at least one child
        child = root * 2 + 1       # Left child
        swap = root                # Keeps track of child to swap with

        if a[swap] < a[child]:
            swap = child
        # If there is a right child and that child is greater
        if child + 1 <= end and a[swap] < a[child + 1]:
            swap = child + 1
        if swap == root:
            # The root holds the largest element. Since we assume the heaps rooted at the
            # children are valid, this means that we are done.
            print(a)
            tree_utils.treePrint(a[:end+1])
            print()
            return
        else:
            a[root], a[swap] = a[swap], a[root]            
            print('swap', a[root], a[swap], 'getting', a)
            root = swap            # repeat to continue sifting down the child now

    tree_utils.treePrint(a[:end+1])


###############################################################################################################
# Put elements of 'a' in heap order, in-place
#
def heapify(a):
    print(a)
    tree_utils.treePrint(a)
    count = len(a)
    # start is assigned the index in 'a' of the last parent node
    # the last element in a 0-based array is at index count-1; find the parent of that element
    start = (count - 2) // 2

    while start >= 0:
        # sift down the node at index 'start' to the proper place such that all nodes below
        # the start index are in heap order
        siftDown(a, start, count - 1)
        
        # go to the next parent node)
        start -= 1
    # after sifting down the root all nodes/elements are in heap order
    

###############################################################################################################
# input: an unordered array a
#    
#    
def heapsort(a):
    count = len(a)

    # Build the heap in array a so that largest value is at the root
    print ('HEAPIFY\n')
    heapify(a)

    # The following loop maintains the invariants that a[0:end] is a heap and every element
    # beyond end is greater than everything before it (so a[end:count] is in sorted order
    end = count - 1

    print ('\nSORT')
    while end > 0:
        print('---------------------------------')
        # a[0] is the root and largest value. The swap moves it in front of the sorted elements.
        print('swap',a[0], a[end], 'in', a, 'getting')
        a[0], a[end] = a[end], a[0]
        tree_utils.treePrint(a[:end])
        print('with sorted:', a[end:])
        # the heap size is reduced by one
        end = end - 1
        # the swap ruined the heap property, so restore it
        siftDown(a, 0, end)



t = [54, 26, 93, 44, 77, 31, 18, 55, 9]
#t = [6, 5, 3, 1, 8, 7, 2, 4]
heapsort(t)
print('\nSORTED\n', t)

#t = [55, 44, 54, 9, 26, 31, 18, 77, 93]
#t = [54, 26, 93, 44, 77, 31, 18, 55, 9]
#
#print(t, '\n')
#heapify(t)
#tree_utils.treePrint(t)
#print()
#print(t)
