theAlphabet = {'G', 'A', 'C', 'T'}

def basicStringSearch(searchString, target):
    searchIndex = 0
    while searchIndex + len(target) <= len(searchString):
        targetIndex = 0
        while (targetIndex < len(target)) and target[targetIndex] == searchString[targetIndex + searchIndex]:
            targetIndex = targetIndex + 1
        if targetIndex == len(target):
            return searchIndex
        searchIndex = searchIndex + 1
    return -1


def stringSearch(searchSpace, target):
    fchar = target[0]                 # first character of target
    tlen  = len(target)
    slast = len(searchSpace)- tlen    # last searchable index
    searchIndex = 0
    while searchIndex <= slast:
        if searchSpace[searchIndex] == fchar:
            targetIndex = 1
            while targetIndex < tlen and target[targetIndex] == searchSpace[targetIndex + searchIndex]:
                targetIndex += 1
            if targetIndex == tlen:
                return searchIndex
        searchIndex += 1
    return -1


# Helper function for quickSearch()
def buildShiftTable(target, alphabet):
    shiftTable = {}
    for char in alphabet:
        shiftTable[char] = len(target) + 1
    for i in range(len(target)):
        character = target[i]
        shift = len(target) - i
        shiftTable[character] = shift
    return shiftTable

def optQuickSearch(searchString, target, alphabet = theAlphabet):

    length_search_string = len(searchString)
    length_target = len(target)
    shiftTable = buildShiftTable(target, alphabet)
    last_index = length_search_string - length_target   # last searchable index
    first_char = target[0]                              # first character of target

    searchIndex = 0
    while searchIndex <= last_index:
        # is the target in the block [searchIndex] .. [searchIndex + length_target-1]
        if searchString[searchIndex] == first_char:
            targetIndex = 1
            while targetIndex < length_target and target[targetIndex] == searchString[searchIndex+targetIndex]:
                targetIndex += 1
            if targetIndex == length_target:
                return searchIndex

        # target not in current block so try next viable block
        skip = searchIndex + length_target
        if skip >= length_search_string:
            return -1
        next = searchString[skip]
        searchIndex = searchIndex + shiftTable[next]
    return -1

def fullQuickSearch(searchString, target, alphabet = theAlphabet):

    length_search_string = len(searchString)
    length_target = len(target)
    shiftTable = buildShiftTable(target, alphabet)
    last_index = length_search_string - length_target   # last searchable index
    first_char = target[0]                              # first character of target

    searchIndex = 0
    if searchString[searchIndex] == first_char:
        targetIndex = 1
        while targetIndex < length_target and target[targetIndex] == searchString[searchIndex+targetIndex]:
            targetIndex += 1
        if targetIndex == length_target:
            return searchIndex
    skip = searchIndex + length_target

    while skip <= last_index:
        next = searchString[skip]
        searchIndex = searchIndex + shiftTable[next]
        # is the target in the block [searchIndex] .. [searchIndex + length_target-1]
        if searchString[searchIndex] == first_char:
            targetIndex = 1
            while targetIndex < length_target and target[targetIndex] == searchString[searchIndex+targetIndex]:
                targetIndex += 1
            if targetIndex == length_target:
                return searchIndex

        # target not in current block so try next viable block
        skip = searchIndex + length_target

    if skip < length_target:
        next = searchString[skip]
        searchIndex = searchIndex + shiftTable[next]

    while searchIndex <= last_index:
        # is the target in the block [searchIndex] .. [searchIndex + length_target-1]
        if searchString[searchIndex] == first_char:
            targetIndex = 1
            while targetIndex < length_target and target[targetIndex] == searchString[searchIndex+targetIndex]:
                targetIndex += 1
            if targetIndex == length_target:
                return searchIndex

        # target not in current block so try next viable block
        skip = searchIndex + length_target
        if skip >= length_search_string:
            return -1
        next = searchString[skip]
        searchIndex = searchIndex + shiftTable[next]
    return -1

def quickSearch(searchString, target, alphabet = theAlphabet):
    shiftTable = buildShiftTable(target, alphabet)
    searchIndex = 0
    while searchIndex + len(target) <= len(searchString):
        targetIndex = 0
        while targetIndex < len(target) and target[targetIndex] == searchString[searchIndex+targetIndex]:
            targetIndex = targetIndex + 1
        if targetIndex == len(target):
            return searchIndex
        if searchIndex + len(target) < len(searchString):       #  - 1 removed
            next = searchString[searchIndex + len(target)]
            shift = shiftTable[next]
            searchIndex = searchIndex + shift
        else:
            return -1
    return -1



# Helper function for kmpSearch()
def buildPrefixTable(target):
    prefixTable = [0] * len(target)
    q = 0
    for p in range(1, len(target)):
        while q > 0 and target[q] != target[p]:
            q = prefixTable[q - 1]
        if target[q] == target[p]:
            q = q + 1
        prefixTable[p] = q
    return prefixTable


def kmpSearch(searchString, target):
    n = len(searchString)
    m = len(target)
    prefixTable = buildPrefixTable(target)
    q = 0
    for i in range(n):
        ########################## print('T[q]:', target[q], 'S[i]:', searchString[i], )
        while q > 0 and target[q] != searchString[i]:
            q = prefixTable[q - 1]
            ##################### print('T[q]:', target[q], 'S[i]:', searchString[i], q, 'from prefix')
        if target[q] == searchString[i]:
            q = q + 1
        if q == m:
            return i - m + 1
    return -1



target1 = 'CGTCCA'
target2 = 'TATATA'
target3 = 'CTCTATGTTTAAAGTTCGA'